from ..widget_base import Output
from urllib.parse import quote
from validators import url


class IframeOutput(Output):
    type = "iframe-output"

    def __init__(self, url_or_html: str, **kwargs):
        self.set_props(dict(url_or_html=url_or_html, **kwargs))

    def set_props(self, props):
        url_or_html = props.get("url_or_html", "")
        if url(url_or_html):
            self.url = url_or_html
        else:
            self.url = f"data:text/html,{quote(url_or_html)}"
        self.width = props.get("width", "800")
        self.height = props.get("height", "600")
        self.full_width = props.get("full_width", False)

    def render(self, ctx: dict):
        return {
            "type": self.type,
            "url": self.url,
            "width": self.width,
            "height": self.height,
            "fullWidth": self.full_width,
        }



from unittest import TestCase
from flask import Flask, Blueprint
from ...repositories.members import LocalMembersRepository
from ...server.guards.role_guard import RoleGuardFactory
from ...jwt_auth import USER_AUTH_HEADER_KEY, UserClaims


class TestRequirementsApi(TestCase):
    def test_app(self):
        ##Autorization: Bearer <test_email>
        def test_header(email: str):
            return {USER_AUTH_HEADER_KEY: f"Bearer {email}"}

        def test_auth_decoder(email: str):
            email = email.split(" ")[1]
            return UserClaims(claims={"email": email})

        repo = LocalMembersRepository({"tester@abstra.app": ["workflow_viewer"]})
        guard = RoleGuardFactory(repo, test_auth_decoder)

        app = Flask(__name__)
        bp = Blueprint("dummy", __name__)

        @bp.route("/")
        @guard.requires("workflow_viewer")
        def main():
            return {"message": "OK"}

        app.register_blueprint(bp, url_prefix="/")

        client = app.test_client()
        response = client.get("/", headers=test_header("tester@abstra.app"))

        self.assertEqual(response.status, "200 OK")

        client = app.test_client()
        response = client.get("/", headers=test_header("not-a-tester@abstra.app"))

        self.assertEqual(response.status, "403 FORBIDDEN")



from .stdio_monkey_patch import override_stdio
from .settings import SettingsController


def cloud_setup(root: str):
    SettingsController.set_root_path(root)
    SettingsController.set_server_port(None)
    override_stdio(print_exceptions=True)



import abstra.forms as af
import abstra.workflows as aw

# With Abstra Forms, it's easy to build user interfaces
name = af.read("👋 Hello there! What is your name?")

# You can store data in the current thread,
# so it will be available in the next stages
aw.set_data("name", name)

# Different kinds of input and output widgets are available
af.display(f"🎉 Welcome, {name}!")

af.display_markdown("Check out our [docs](https://docs.abstra.io/forms/overview) 📚")



from abstra_internals.interface.sdk.tables.api import (
    run,
    query_df,
    TablesExecutionError,
    select,
    select_by_id,
    select_df,
    select_one,
    insert,
    update,
    delete,
    update_by_id,
    delete_by_id,
)



import debugpy
from ...utils import get_free_port

CONNECTED_PORT = None


def start_debugger():
    global CONNECTED_PORT
    if CONNECTED_PORT is None:
        CONNECTED_PORT = get_free_port(5678)
        debugpy.listen(CONNECTED_PORT)
    return CONNECTED_PORT


def is_client_connected():
    return debugpy.is_client_connected()



import inspect, typing, traceback
from types import TracebackType
from dataclasses import dataclass

from .settings import Settings

Frames = typing.Union[typing.List[inspect.FrameInfo], Exception]


def representations(locals: typing.Dict):
    result = {}
    for key, value in list(locals.items()):
        try:
            result[key] = repr(value)
        except:
            pass
    return result


def traceback_to_infos(
    traceback: typing.Union[TracebackType, None]
) -> typing.List[inspect.FrameInfo]:
    return inspect.getouterframes(traceback.tb_frame) if traceback else []


def _make_debug_item_from_info(info: inspect.FrameInfo):
    return {
        "filename": info.filename,
        "lineno": info.lineno,
        "name": info.function,
        "locals": representations(info.frame.f_locals),
    }


def _make_debug_item_from_stack(frame_summary: traceback.FrameSummary):
    return {
        "filename": frame_summary.filename,
        "lineno": frame_summary.lineno,
        "name": frame_summary.name,
        "locals": representations(frame_summary.locals or {}),
    }


def make_debug_data(frames_or_exception: Frames):
    if isinstance(frames_or_exception, Exception):
        e = frames_or_exception
        frames = traceback.TracebackException(
            type(e), e, e.__traceback__, capture_locals=True
        ).stack
    else:
        frames = frames_or_exception
    root_path = str(Settings.root_path)
    return {
        "debug": {
            "stack": [
                (
                    _make_debug_item_from_info(info)
                    if isinstance(info, inspect.FrameInfo)
                    else _make_debug_item_from_stack(info)
                )
                for info in (frames)
                if root_path in info.filename
            ]
        }
    }


@dataclass
class CloseDTO:
    exit_code: typing.Union[int, None] = None
    exception: typing.Union[str, None] = None
    frames: typing.Optional[Frames] = None



import typing

from ....widgets.widget_base import Input


class Reactive(Input):
    type = "reactive"

    def __init__(self, callback: typing.Callable, **kwargs):
        self.set_props(dict(callback=callback, **kwargs))
        self.page = None

    def has_errors(self):
        return self.page.has_errors() if self.page else False

    def render(self, ctx: dict):
        payload = {}
        payload.update(self.value)
        payload.update(ctx)
        self.page = self.callback(payload)
        if self.page is None:
            return []

        if self.value:
            self.page.set_values(self.value)
            self.page.set_errors()

        return self.page.render(ctx)

    def set_value(self, value, set_errors=False):
        self.value = value
        if hasattr(self, "page") and self.page:
            self.page.set_values(value)

    def serialize_value(self):
        return self.page.serialize_value() if self.page else {}

    def set_props(self, props):
        self.callback = props.get("callback")
        self.set_value(props.get("initial_value", {}))
        self.page = None

    def set_errors(self):
        if self.page:
            self.page.set_errors()

    def parse_value(self, value):
        if self.page:
            parsed = self.page.parse_value(value)
            return parsed
        else:
            return {}
